@@ -350,11 +350,14 @@ checkOverlap env =
                 pure (forceUnambig env)
 
 -- | Issue warnings if entries in the first environment would
--- shadow something in the second.
+--   shadow something in the second. This warning is only emited
+--   for UserNames
 checkShadowing :: NamingEnv -> NamingEnv -> RenameM ()
 checkShadowing envNew envOld =
-  mapM_ recordWarning
-    [ SymbolShadowed p x xs | (p,x,xs) <- findShadowing envNew envOld ]
+  mapM_
+    recordWarning
+    [ SymbolShadowed p x xs | (p, x, xs) <- findShadowing envNew envOld
+    ]
 
 
 -- | Shadow the current naming environment with some more names.
@@ -525,4 +528,4 @@ isFakeName m =
     ImpNested x ->
       case nameTopModuleMaybe x of
         Just y  -> y == undefinedModName
-        Nothing -> False
+        Nothing -> False
\ No newline at end of file
